/**
 * Copyright (c) 2017-2021 Samsung Electronics Co., Ltd All Rights Reserved
 *
 *    Licensed under the Apache License, Version 2.0 (the "License");
 *    you may not use this file except in compliance with the License.
 *    You may obtain a copy of the License at
 *
 *        http://www.apache.org/licenses/LICENSE-2.0
 *
 *    Unless required by applicable law or agreed to in writing, software
 *    distributed under the License is distributed on an "AS IS" BASIS,
 *    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *    See the License for the specific language governing permissions and
 *    limitations under the License.
 */
/**
 * @file
 * @author Lukasz Kostyra (l.kostyra@samsung.com)
 * @brief  Teec to/from JSON translation function definitions
 */

#include "TeecTranslations.h"
#include "TeecManager.h"
#include "TeecContext.h"
#include "TeecSharedMemory.h"

#pragma GCC diagnostic ignored "-Wdeprecated-declarations"
#include <tee_client_api.h>

#define CASE_RESULT_TO_STRING(x) case x: return #x

namespace {

// JSON TEEC_Parameter parsing tokens
const std::string PARAM_TYPE_KEY = "type";
const std::string PARAM_TYPE_NONE = "NONE";
const std::string PARAM_TYPE_VALUE = "VALUE";
const std::string PARAM_TYPE_MEMREF = "MEMREF";
const std::string PARAM_TYPE_TMPREF = "TMPREF";
const std::string PARAM_FLAG_KEY = "flag";
const std::string PARAM_FLAG_NONE = "NONE";
const std::string PARAM_FLAG_INPUT = "INPUT";
const std::string PARAM_FLAG_OUTPUT = "OUTPUT";
const std::string PARAM_FLAG_INOUT = "INOUT";
const std::string PARAM_FLAG_WHOLE = "WHOLE";
const std::string PARAM_FLAG_PARTIAL_INPUT = "PARTIAL_INPUT";
const std::string PARAM_FLAG_PARTIAL_OUTPUT = "PARTIAL_OUTPUT";
const std::string PARAM_FLAG_PARTIAL_INOUT = "PARTIAL_INOUT";
const std::string PARAM_VALUE_A = "a";
const std::string PARAM_VALUE_B = "a";
const std::string PARAM_MEMREF_SIZE = "size";
const std::string PARAM_MEMREF_OFFSET = "offset";
const std::string PARAM_MEMREF_SHM = "shm";
const std::string PARAM_TMPREF_MEM = "mem";

const uint32_t TEEC_OPERATION_COUNT = 4;

} // namespace

namespace extension {
namespace libteec {
namespace translations {

std::string TeecResultToString(TEEC_Result result) {
    switch (result) {
    CASE_RESULT_TO_STRING(TEEC_SUCCESS);
    CASE_RESULT_TO_STRING(TEEC_ERROR_GENERIC);
    CASE_RESULT_TO_STRING(TEEC_ERROR_ACCESS_DENIED);
    CASE_RESULT_TO_STRING(TEEC_ERROR_CANCEL);
    CASE_RESULT_TO_STRING(TEEC_ERROR_ACCESS_CONFLICT);
    CASE_RESULT_TO_STRING(TEEC_ERROR_EXCESS_DATA);
    CASE_RESULT_TO_STRING(TEEC_ERROR_BAD_FORMAT);
    CASE_RESULT_TO_STRING(TEEC_ERROR_BAD_PARAMETERS);
    CASE_RESULT_TO_STRING(TEEC_ERROR_BAD_STATE);
    CASE_RESULT_TO_STRING(TEEC_ERROR_ITEM_NOT_FOUND);
    CASE_RESULT_TO_STRING(TEEC_ERROR_NOT_IMPLEMENTED);
    CASE_RESULT_TO_STRING(TEEC_ERROR_NOT_SUPPORTED);
    CASE_RESULT_TO_STRING(TEEC_ERROR_NO_DATA);
    CASE_RESULT_TO_STRING(TEEC_ERROR_OUT_OF_MEMORY);
    CASE_RESULT_TO_STRING(TEEC_ERROR_BUSY);
    CASE_RESULT_TO_STRING(TEEC_ERROR_COMMUNICATION);
    CASE_RESULT_TO_STRING(TEEC_ERROR_SECURITY);
    CASE_RESULT_TO_STRING(TEEC_ERROR_SHORT_BUFFER);
    CASE_RESULT_TO_STRING(TEEC_ERROR_EXTERNAL_CANCEL);
    CASE_RESULT_TO_STRING(TEEC_ERROR_TARGET_DEAD);
    // not throwing here, there are TEE internal error not known from
    // TEEC header, moreover this function is usually used in throws anyway
    default: return "TEEC_ERROR_UNKNOWN";
    }
}

uint32_t StringToTeecLoginMethod(const std::string& str) {
    if (str == "PUBLIC") return TEEC_LOGIN_PUBLIC;
    if (str == "USER") return TEEC_LOGIN_USER;
    if (str == "GROUP") return TEEC_LOGIN_GROUP;
    if (str == "APPLICATION") return TEEC_LOGIN_APPLICATION;
    throw common::InvalidValuesException("Invalid login type: " + str);
}

uint32_t ParamTypeFromJSON(const picojson::object& obj) {
    auto typeVal = obj.find(PARAM_TYPE_KEY);
    if (typeVal == obj.end()) {
        throw common::InvalidValuesException("Missing parameter type info");
    }

    auto flagVal = obj.find(PARAM_FLAG_KEY);
    if (flagVal == obj.end()) {
        throw common::InvalidValuesException("Flag parameter not added to operation");
    }

    const std::string& typeStr = typeVal->second.get<std::string>();
    const std::string& flagStr = flagVal->second.get<std::string>();

    if (typeStr == PARAM_TYPE_VALUE) {
        if (flagStr == PARAM_FLAG_INPUT) {
            return TEEC_VALUE_INPUT;
        } else if (flagStr == PARAM_FLAG_OUTPUT) {
            return TEEC_VALUE_OUTPUT;
        } else if (flagStr == PARAM_FLAG_INOUT) {
            return TEEC_VALUE_INOUT;
        } else {
            throw common::InvalidValuesException("Invalid parameter flag info: "
                                                 + flagStr);
        }
    } else if (typeStr == PARAM_TYPE_MEMREF) {
        if (flagStr == PARAM_FLAG_WHOLE) {
            return TEEC_MEMREF_WHOLE;
        } else if (flagStr == PARAM_FLAG_PARTIAL_INPUT) {
            return TEEC_MEMREF_PARTIAL_INPUT;
        } else if (flagStr == PARAM_FLAG_PARTIAL_OUTPUT) {
            return TEEC_MEMREF_PARTIAL_OUTPUT;
        } else if (flagStr == PARAM_FLAG_PARTIAL_INOUT) {
            return TEEC_MEMREF_PARTIAL_INOUT;
        } else {
            throw common::InvalidValuesException("Invalid parameter flag info: "
                                                 + flagStr);
        }
    } else if (typeStr == PARAM_TYPE_TMPREF) {
        if (flagStr == PARAM_FLAG_INPUT) {
            return TEEC_MEMREF_TEMP_INPUT;
        } else if (flagStr == PARAM_FLAG_OUTPUT) {
            return TEEC_MEMREF_TEMP_OUTPUT;
        } else if (flagStr == PARAM_FLAG_INOUT) {
            return TEEC_MEMREF_TEMP_INOUT;
        } else {
            throw common::InvalidValuesException("Invalid parameter flag info: "
                                                 + flagStr);
        }
    }

    throw common::InvalidValuesException("Invalid parameter type info: "
                                         + typeVal->second.get<std::string>());
}

TEEC_SharedMemory* SharedMemoryFromJSON(const picojson::object& object) {
    auto cID = object.find("contextId");
    auto mID = object.find("memId");

    if (cID == object.end() || mID == object.end()) {
        throw common::InvalidValuesException("Invalid shared memory object provided as param");
    }

    const std::string& contextId = cID->second.get<std::string>();
    const std::string& memId = mID->second.get<std::string>();
    TeecContextPtr context = TeecManager::Instance().GetContext(contextId);
    TeecSharedMemoryPtr mem = context->GetSharedMemory(memId);
    return mem->GetMemoryPointer();
}

TEEC_Operation OperationFromJSON(const picojson::value& value, TeecTempMemoryAllocator& allocator) {
    TEEC_Operation result;
    result.started = 0;
    result.paramTypes = 0;

    if (!value.is<picojson::array>()) {
        return result;
    }

    std::vector<uint32_t> paramTypes(TEEC_OPERATION_COUNT);
    for (auto& p: paramTypes) {
        p = TEEC_NONE;
    }

    const picojson::array& array = value.get<picojson::array>();
    if (array.size() > TEEC_OPERATION_COUNT) {
        throw common::InvalidValuesException("Too many operation parameters");
    }

    for (uint32_t i = 0; i < array.size(); ++i) {
        const picojson::object& paramObject = array[i].get<picojson::object>();

        auto type = paramObject.find(PARAM_TYPE_KEY);
        if (type != paramObject.end()) {
            paramTypes[i] = ParamTypeFromJSON(paramObject);
            const std::string& paramTypeStr = type->second.get<std::string>();
            if (paramTypeStr == PARAM_TYPE_VALUE) {
                auto valA = paramObject.find(PARAM_VALUE_A);
                auto valB = paramObject.find(PARAM_VALUE_B);
                if (valA != paramObject.end() && valB != paramObject.end()) {
                    result.params[i].value.a = static_cast<uint32_t>(valA->second.get<double>());
                    result.params[i].value.b = static_cast<uint32_t>(valB->second.get<double>());
                }
            } else if (paramTypeStr == PARAM_TYPE_MEMREF) {
                auto size = paramObject.find(PARAM_MEMREF_SIZE);
                auto offset = paramObject.find(PARAM_MEMREF_OFFSET);
                auto shm = paramObject.find(PARAM_MEMREF_SHM);
                if (size != paramObject.end() &&
                    offset != paramObject.end() &&
                    shm != paramObject.end()) {
                    result.params[i].memref.size = static_cast<size_t>(size->second.get<double>());
                    result.params[i].memref.offset = static_cast<size_t>(offset->second.get<double>());
                    result.params[i].memref.parent = SharedMemoryFromJSON(shm->second.get<picojson::object>());
                }
            } else if (paramTypeStr == PARAM_TYPE_TMPREF) {
                auto mem = paramObject.find(PARAM_TMPREF_MEM);
                if (mem != paramObject.end()) {
                    const picojson::array& tmpMem = mem->second.get<picojson::array>();
                    result.params[i].tmpref.size = tmpMem.size();
                    result.params[i].tmpref.buffer = allocator.AllocateTempMemory(tmpMem);
                }
            } else {
                throw common::InvalidValuesException("Invalid parameter type provided: "
                                                     + type->second.get<std::string>());
            }
        }
    }

    result.paramTypes = TEEC_PARAM_TYPES(paramTypes[0], paramTypes[1], paramTypes[2], paramTypes[3]);
    return result;
}

uint32_t SharedMemoryFlagsFromJSON(const picojson::value& value) {
    if (!value.is<std::string>()) {
        throw common::InvalidValuesException("Attempted to acquire shared memory flag from non-string object");
    }

    const std::string& flagStr = value.get<std::string>();
    if (flagStr == PARAM_FLAG_INPUT) {
        return TEEC_MEM_INPUT;
    } else if (flagStr == PARAM_FLAG_OUTPUT) {
        return TEEC_MEM_OUTPUT;
    } else if (flagStr == PARAM_FLAG_INOUT) {
        return TEEC_MEM_INPUT | TEEC_MEM_OUTPUT;
    }

    throw common::InvalidValuesException("Invalid flag provided: " + flagStr);
}

DataBuffer DataBufferFromJSON(const picojson::value& value) {
    if (!value.is<picojson::array>()) {
        throw common::InvalidValuesException("Attempted to acquire data array from non-array object");
    }

    const picojson::array& array = value.get<picojson::array>();
    DataBuffer buffer(array.size());

    for (size_t i = 0; i < array.size(); ++i) {
        buffer[i] = static_cast<uint8_t>(array[i].get<double>());
    }

    return buffer;
}

std::string JSONFromParamType(uint32_t paramType) {
    switch (paramType) {
    case TEEC_NONE:
        return PARAM_TYPE_NONE;
    case TEEC_VALUE_INPUT:
    case TEEC_VALUE_OUTPUT:
    case TEEC_VALUE_INOUT:
        return PARAM_TYPE_VALUE;
    case TEEC_MEMREF_TEMP_INPUT:
    case TEEC_MEMREF_TEMP_OUTPUT:
    case TEEC_MEMREF_TEMP_INOUT:
        return PARAM_TYPE_TMPREF;
    case TEEC_MEMREF_WHOLE:
    case TEEC_MEMREF_PARTIAL_INPUT:
    case TEEC_MEMREF_PARTIAL_OUTPUT:
    case TEEC_MEMREF_PARTIAL_INOUT:
        return PARAM_TYPE_MEMREF;
    default:
        throw common::InvalidValuesException("Unknown TEEC parameter type: " + std::to_string(paramType));
    };
}

std::string JSONFromParamFlag(uint32_t paramType) {
    switch (paramType) {
    case TEEC_NONE:
        return PARAM_FLAG_NONE;
    case TEEC_VALUE_INPUT:
    case TEEC_MEMREF_TEMP_INPUT:
        return PARAM_FLAG_INPUT;
    case TEEC_VALUE_OUTPUT:
    case TEEC_MEMREF_TEMP_OUTPUT:
        return PARAM_FLAG_OUTPUT;
    case TEEC_VALUE_INOUT:
    case TEEC_MEMREF_TEMP_INOUT:
        return PARAM_FLAG_INOUT;
    case TEEC_MEMREF_WHOLE:
        return PARAM_FLAG_WHOLE;
    case TEEC_MEMREF_PARTIAL_INPUT:
        return PARAM_FLAG_PARTIAL_INPUT;
    case TEEC_MEMREF_PARTIAL_OUTPUT:
        return PARAM_FLAG_PARTIAL_OUTPUT;
    case TEEC_MEMREF_PARTIAL_INOUT:
        return PARAM_FLAG_PARTIAL_INOUT;
    default:
        throw common::InvalidValuesException("Unknown TEEC parameter type: " + std::to_string(paramType));
    };
}

picojson::value UpdateJSONParams(const picojson::value& json, TEEC_Operation op) {
    const picojson::array& inParamArray = json.get<picojson::array>();
    picojson::array outParamArray;

    uint32_t paramTypes[4] = {
        TEEC_PARAM_TYPE_GET(op.paramTypes, 0),
        TEEC_PARAM_TYPE_GET(op.paramTypes, 1),
        TEEC_PARAM_TYPE_GET(op.paramTypes, 2),
        TEEC_PARAM_TYPE_GET(op.paramTypes, 3),
    };

    for (uint32_t i = 0; i < 4; ++i) {
        picojson::object param;

        std::string paramTypeStr = JSONFromParamType(paramTypes[i]);
        param.insert(std::make_pair(
            PARAM_TYPE_KEY, picojson::value(paramTypeStr)));
        param.insert(std::make_pair(
            PARAM_FLAG_KEY, picojson::value(JSONFromParamFlag(paramTypes[i]))));

        if (paramTypeStr == PARAM_TYPE_VALUE) {
            // value type can be completely overwritten
            param.insert(std::make_pair(PARAM_VALUE_A,
                picojson::value(static_cast<double>(op.params[i].value.a))));
            param.insert(std::make_pair(PARAM_VALUE_B,
                picojson::value(static_cast<double>(op.params[i].value.b))));
        } else if (paramTypeStr == PARAM_TYPE_MEMREF) {
            const picojson::object& memref = inParamArray[i].get<picojson::object>();
            auto shm = memref.find("shm");
            if (shm == memref.end()) {
                throw common::InvalidValuesException("Invalid shm object");
            }

            // memref type copies shm from source, rest parameters from TA
            param.insert(std::make_pair(PARAM_MEMREF_SIZE,
                picojson::value(static_cast<double>(op.params[i].memref.size))));
            param.insert(std::make_pair(PARAM_MEMREF_OFFSET,
                picojson::value(static_cast<double>(op.params[i].memref.offset))));
            param.insert(std::make_pair(PARAM_MEMREF_SHM, shm->second));
        } else if (paramTypeStr == PARAM_TYPE_TMPREF) {
            picojson::array buf(op.params[i].tmpref.size);
            uint8_t* opBuf = reinterpret_cast<uint8_t*>(op.params[i].tmpref.buffer);
            for (size_t i = 0; i < buf.size(); ++i)
                buf[i] = picojson::value(static_cast<double>(opBuf[i]));

            param.insert(std::make_pair(PARAM_TMPREF_MEM, picojson::value(buf)));
        } else if (paramTypeStr != PARAM_TYPE_NONE) {
            throw common::InvalidValuesException(
                "Invalid parameter type read from operation: " + paramTypeStr);
        }

        outParamArray.push_back(picojson::value(param));
    }

    return picojson::value(outParamArray);
}

picojson::value JSONFromDataBuffer(const DataBuffer& buffer) {
    picojson::array result(buffer.size());

    for (uint32_t i = 0; i < buffer.size(); ++i) {
        result[i] = picojson::value(static_cast<double>(buffer[i]));
    }

    return picojson::value(result);
}

} // namespace translations
} // namespace libteec
} // namespace extension
